# module to hold the function to make our fairly complicated angular restraint string expression for CustomCompoundBondForce in OpenMM

# import the OpenMM stuff
from simtk.openmm.app import *
from simtk.openmm import *
from simtk.unit import *

# import the math library
import math

# function to create the actual restraint string
def CreateAngleRestraintString(g1,g2,g3,g1m,g2m,g3m,option1,option2):

	if option1 == 'COG':
	        print 'Creating Centre of Geometry Angle Restraint Energy Expression...'

		# setup our common characters for string insertion
		x = 'x'
		y = 'y'
		z = 'z'

		# get the center of the first point
		x1 = '('
		y1 = '('
		z1 = '('	
        	for i in range(0,len(g1)):
            		if i != 0:
                		x1 += ' + '
                		y1 += ' + '
                		z1 += ' + '
            		k = str(i+1)
            		x1 += x + k
            		y1 += y + k
            		z1 += z + k

        	x1 += ')'
        	y1 += ')'
        	z1 += ')'
        	M1 = str(1.0 / float(len(g1)))

		# get the center of the second point
		x2 = '('
		y2 = '('
		z2 = '('	
        	for i in range(len(g1),len(g1)+len(g2)):
            		if i != len(g1):
                		x2 += ' + '
                		y2 += ' + '
                		z2 += ' + '
            		k = str(i+1)
            		x2 += x + k
            		y2 += y + k
            		z2 += z + k

        	x2 += ')'
        	y2 += ')'
        	z2 += ')'
        	M2 = str(1.0 / float(len(g2)))

		# get the center of the third point
		x3 = '('
		y3 = '('
		z3 = '('	
        	for i in range(len(g1)+len(g2),len(g1)+len(g2)+len(g3)):
            		if i != len(g1)+len(g2):
                		x3 += ' + '
                		y3 += ' + '
                		z3 += ' + '
            		k = str(i+1)
            		x3 += x + k
            		y3 += y + k
            		z3 += z + k

        	x3 += ')'
        	y3 += ')'
        	z3 += ')'
        	M3 = str(1.0 / float(len(g3)))

    	elif option1 == 'COM':
        	print 'Creating Centre of Mass Angle Restraint Energy Expression...'

		# setup our common characters for string insertion
		x = 'x'
		y = 'y'
		z = 'z'

		# get the center of the first point
		x1 = '('
		y1 = '('
		z1 = '('	

        	for i in range(0,len(g1)):
            		if i != 0:
                		x1 += ' + '
                		y1 += ' + '
                		z1 += ' + '
				M1 += ' + '
            		k = str(i+1)
            		x1 += str(g1m[i].value_in_unit(dalton)) + '*' + x + k
            		y1 += str(g1m[i].value_in_unit(dalton)) + '*' + y + k
            		z1 += str(g1m[i].value_in_unit(dalton)) + '*' + z + k

        	x1 += ')'
        	y1 += ')'
        	z1 += ')'

		# get the center of the second point
		x2 = '('
		y2 = '('
		z2 = '('	

        	for i in range(len(g1),len(g1)+len(g2)):
            		if i != len(g1):
                		x2 += ' + '
                		y2 += ' + '
                		z2 += ' + '
            		k = str(i+1)
            		x2 += str(g2m[i - len(g1)].value_in_unit(dalton)) + '*' + x + k
            		y2 += str(g2m[i - len(g1)].value_in_unit(dalton)) + '*' + y + k
            		z2 += str(g2m[i - len(g1)].value_in_unit(dalton)) + '*' + z + k

        	x2 += ')'
        	y2 += ')'
        	z2 += ')'

		# get the center of the third point
		x3 = '('
		y3 = '('
		z3 = '('
	
        	for i in range(len(g1)+len(g2),len(g1)+len(g2)+len(g3)):
            		if i != len(g1)+len(g2):
                		x3 += ' + '
                		y3 += ' + '
                		z3 += ' + '
            		k = str(i+1)
            		x3 += str(g3m[i - (len(g1) + len(g2))].value_in_unit(dalton)) + '*' + x + k
            		y3 += str(g3m[i - (len(g1) + len(g2))].value_in_unit(dalton)) + '*' + y + k
            		z3 += str(g3m[i - (len(g1) + len(g2))].value_in_unit(dalton)) + '*' + z + k

        	x3 += ')'
        	y3 += ')'
        	z3 += ')'

		# sum up the masses so we can be more efficient with an invert and multiply
		M1Sum = 0.0
		M2Sum = 0.0
		M3Sum = 0.0
		for mass in g1m:
			M1Sum += mass.value_in_unit(dalton)
		for mass in g2m:
			M2Sum += mass.value_in_unit(dalton)
		for mass in g3m:
			M3Sum += mass.value_in_unit(dalton)

		M1 = '(' + str(1.0 / M1Sum) + ')'
		M2 = '(' + str(1.0 / M2Sum) + ')'
		M3 = '(' + str(1.0 / M3Sum) + ')'

	# the remainder of the setup now uses the COM or COG generated centers
	# calculate the weighted the centers
	x1 = '(' + x1 + '*' + M1 + ')'
	y1 = '(' + y1 + '*' + M1 + ')'
	z1 = '(' + z1 + '*' + M1 + ')'
	x2 = '(' + x2 + '*' + M2 + ')'
	y2 = '(' + y2 + '*' + M2 + ')'
	z2 = '(' + z2 + '*' + M2 + ')'
	x3 = '(' + x3 + '*' + M3 + ')'
	y3 = '(' + y3 + '*' + M3 + ')'
	z3 = '(' + z3 + '*' + M3 + ')'

	# now do the calculations using these new centers as the points
	# calculate the first vector
	v0x = '(' + x2 + '-' + x1 + ')' 
	v0y = '(' + y2 + '-' + y1 + ')' 
	v0z = '(' + z2 + '-' + z1 + ')' 
	# calculate the second vector
	v1x = '(' + x2 + '-' + x3 + ')' 
	v1y = '(' + y2 + '-' + y3 + ')' 
	v1z = '(' + z2 + '-' + z3 + ')' 
	# calculate the normalized magnitude of the vectors and get the inverse of their product
	r0 = 'sqrt(' + v0x + '*' + v0x + '+' + v0y + '*' + v0y + '+' + v0z + '*' + v0z + ')'
	r1 = 'sqrt(' + v1x + '*' + v1x + '+' + v1y + '*' + v1y + '+' + v1z + '*' + v1z + ')'
	r0r1 = '(1.0 / (' + r0 + '*' + r1 + '))'
	# calculate the dot product of the vectors
	dot = '(' + v0x + '*' + v1x + '+' + v0y + '*' + v1y + '+' + v0z + '*' + v1z + ')'

	# calculate the cosine of the angle between the vectors, but adjust for the vector magnitude
	cosine = '(' + dot + '*' + r0r1 + ')'

	# calculate theta using arccosine, which will be good for all values not near 0 or 180
	#angle1 = '(acos(' + cosine + '))'
	angle1 = '(acos(min(max(' + cosine + ', -1), 1)))'	
	#angle1 = '(acos(min(max(' + cosine + ', -0.99), 0.99)))'	# uncomment this line and the similar one below to make it work properly

	# do another calculation of theta using the asin method
	# calculate the cross product of the vectors
	c0x = '((' + v0z + '*' + v1y + ')' + ' - ' + '(' + v0y + '*' + v1z + '))'
	c0y = '((' + v0x + '*' + v1z + ')' + ' - ' + '(' + v0z + '*' + v1x + '))'
	c0z = '((' + v0y + '*' + v1x + ')' + ' - ' + '(' + v0x + '*' + v1y + '))'	

	# square each vector and multliply the products
	scale = '((' + v0x + '*' + v0x + '+' + v0y + '*' + v0y + '+' + v0z + '*' + v0z + ')' + '*' + '(' + v1x + '*' + v1x + '+' + v1y + '*' + v1y + '+' + v1z + '*' + v1z + '))'

	# square the cross product
	squareCross = '(' + c0x + '*' + c0x + '+' + c0y + '*' + c0y + '+' + c0z + '*' + c0z + ')'

	# calculate theta using arcsine (from P.Eastman's code)
	#angle2raw = 'asin(' + 'sqrt(' + squareCross + '/' + scale + '))'
	angle2raw = 'asin(min(max(' + '(sqrt(' + squareCross + '/' + scale + ')), -1), 1))'	
	#angle2raw = 'asin(min(max(' + '(sqrt(' + squareCross + '/' + scale + ')), -0.99), 0.99))'	# uncomment this line and the similar one above to make it work properly

	# fix the angle to be in the right range (from P.Eastman's code)
	angle2 = '((step(' + cosine + ')' + '*' + angle2raw + ')' + '+' + '((' + '1' + '-' + 'step(' + cosine + '))' + '*' + '(' + '3.14159265' + '-' + angle2raw  + ')' + '))'

	# use a step function to select the appropriate method for calculating the angle.  step(x) returns 0 if x is less than 0, otherwise 1
	select = '(step(abs(' + cosine + ') - 0.99))'	# this will return 1 when cosine needs to be calculated by the asin method, 0 otherwise
	angle = '((' + select + '*' + angle2 + ')' + '+' + '((' + '1' + '-' + select + ')' + '*' + angle1 + '))'

	# allow the user to select the method being used
	#if option2 == 'full':
	#	return_string = '0.5 * ' + 'angleK' + ' * ' + '(' + 'theta0' + ' - ' + angle + ')^2'
	#elif option2 == 'acos':
	#	return_string = '0.5 * ' + 'angleK' + ' * ' + '(' + 'theta0' + ' - ' + angle1 + ')^2'
	#elif option2 == 'asin':
	#	return_string = '0.5 * ' + 'angleK' + ' * ' + '(' + 'theta0' + ' - ' + angle2 + ')^2'

	return_string = '0.5 * ' + 'angleK' + ' * ' + '(' + 'theta0' + ' - ' + angle + ')^2'

	return return_string	


# function to calculate the current angle using the same inputs as the restraint string
def CalculateAngle(Positions, g1, g2, g3, g1m, g2m, g3m, option):

	x1 = 0.0
	y1 = 0.0
	z1 = 0.0
	x2 = 0.0
	y2 = 0.0
	z2 = 0.0
	x3 = 0.0
	y3 = 0.0
	z3 = 0.0
	M1 = 0.0
	M2 = 0.0
	M3 = 0.0

	try:
		if option == 'COM':	
			for count, index in enumerate(g1):
				x1 += Positions[index][0].value_in_unit(nanometer) * g1m[count].value_in_unit(dalton)
				y1 += Positions[index][1].value_in_unit(nanometer) * g1m[count].value_in_unit(dalton)
				z1 += Positions[index][2].value_in_unit(nanometer) * g1m[count].value_in_unit(dalton)
				M1 += g1m[count].value_in_unit(dalton)
			for count, index in enumerate(g2):
				x2 += Positions[index][0].value_in_unit(nanometer) * g2m[count].value_in_unit(dalton)
				y2 += Positions[index][1].value_in_unit(nanometer) * g2m[count].value_in_unit(dalton)
				z2 += Positions[index][2].value_in_unit(nanometer) * g2m[count].value_in_unit(dalton)
				M2 += g2m[count].value_in_unit(dalton)
			for count, index in enumerate(g3):
				x3 += Positions[index][0].value_in_unit(nanometer) * g3m[count].value_in_unit(dalton)
				y3 += Positions[index][1].value_in_unit(nanometer) * g3m[count].value_in_unit(dalton)
				z3 += Positions[index][2].value_in_unit(nanometer) * g3m[count].value_in_unit(dalton)
				M3 += g3m[count].value_in_unit(dalton)

		elif option == 'COG':
			for count, index in enumerate(g1):
				x1 += Positions[index][0].value_in_unit(nanometer)
				y1 += Positions[index][1].value_in_unit(nanometer)
				z1 += Positions[index][2].value_in_unit(nanometer)
				M1 += 1.0
			for count, index in enumerate(g2):
				x2 += Positions[index][0].value_in_unit(nanometer)
				y2 += Positions[index][1].value_in_unit(nanometer)
				z2 += Positions[index][2].value_in_unit(nanometer)
				M2 += 1.0
			for count, index in enumerate(g3):
				x3 += Positions[index][0].value_in_unit(nanometer)
				y3 += Positions[index][1].value_in_unit(nanometer)
				z3 += Positions[index][2].value_in_unit(nanometer)
				M3 += 1.0
	except AttributeError:
		if option == 'COM':	
			for count, index in enumerate(g1):
				x1 += Positions[index][0] * g1m[count].value_in_unit(dalton)
				y1 += Positions[index][1] * g1m[count].value_in_unit(dalton)
				z1 += Positions[index][2] * g1m[count].value_in_unit(dalton)
				M1 += g1m[count].value_in_unit(dalton)
			for count, index in enumerate(g2):
				x2 += Positions[index][0] * g2m[count].value_in_unit(dalton)
				y2 += Positions[index][1] * g2m[count].value_in_unit(dalton)
				z2 += Positions[index][2] * g2m[count].value_in_unit(dalton)
				M2 += g2m[count].value_in_unit(dalton)
			for count, index in enumerate(g3):
				x3 += Positions[index][0] * g3m[count].value_in_unit(dalton)
				y3 += Positions[index][1] * g3m[count].value_in_unit(dalton)
				z3 += Positions[index][2] * g3m[count].value_in_unit(dalton)
				M3 += g3m[count].value_in_unit(dalton)

		elif option == 'COG':
			for count, index in enumerate(g1):
				x1 += Positions[index][0]
				y1 += Positions[index][1]
				z1 += Positions[index][2]
				M1 += 1.0
			for count, index in enumerate(g2):
				x2 += Positions[index][0]
				y2 += Positions[index][1]
				z2 += Positions[index][2]
				M2 += 1.0
			for count, index in enumerate(g3):
				x3 += Positions[index][0]
				y3 += Positions[index][1]
				z3 += Positions[index][2]
				M3 += 1.0
	
	# calculate the geometry weighted centers
	x1 = x1 / M1
	y1 = y1 / M1
	z1 = z1 / M1
	x2 = x2 / M2
	y2 = y2 / M2
	z2 = z2 / M2
	x3 = x3 / M3
	y3 = y3 / M3
	z3 = z3 / M3
		
	# calculate the two vectors that make up the angle
	# calculate the first vector
	v0x = x2 - x1
	v0y = y2 - y1
	v0z = z2 - z1
	# calculate the second vector
	v1x = x2 - x3
	v1y = y2 - y3
	v1z = z2 - z3
	# calculate the normalized magnitude of the vectors and get the inverse of their product
	r0 = math.sqrt( v0x * v0x + v0y * v0y + v0z * v0z )
	r1 = math.sqrt( v1x * v1x + v1y * v1y + v1z * v1z )
	r0r1 = (1.0 / ( r0 * r1 ))
	# calculate the dot product of the vectors
	dot = v0x * v1x + v0y * v1y + v0z * v1z

	# calculate the cosine of the angle between the vectors, but ensure it is within the reasonable range of 1.0 to -1.0
	cosine = dot * r0r1
	# calculate theta, adding some extra bits to make sure the return value makes sense, but NaN errors can still happen, likely because of the gradients which I can't control
	angle = math.acos( cosine )

	angle = angle * (360.0 / (2.0 * math.pi))

	return angle

